"""

.. _Condition_Overview:

Overview
--------

:class:`Conditions <Condition>` are used to specify when nodes are allowed to execute.  Conditions
can be used to specify a variety of required conditions for execution, including the state of the node
itself (e.g., how many times it has already executed, or the value of one of its attributes), the state of the
:class:`Scheduler <graph_scheduler.scheduler.Scheduler>` (e.g., how many `CONSIDERATION_SET_EXECUTION` s have occurred in the current `ENVIRONMENT_STATE_UPDATE`), or the state of other
nodes in a graph (e.g., whether or how many times they have executed). This package provides a number of
`pre-specified Conditions <Condition_Pre_Specified>` that can be parametrized (e.g., how many times a node should
be executed). `Custom conditions <Condition_Custom>` can also be created, by assigning a function to a Condition that
can reference any node or its attributes, thus providing considerable flexibility for scheduling.

.. _Condition_Creation:

Creating Conditions
-------------------

.. _Condition_Pre_Specified:

*Pre-specified Conditions*
~~~~~~~~~~~~~~~~~~~~~~~~~~

`Pre-specified Conditions <Condition_Pre-Specified_List>` can be instantiated and added to a :class:`Scheduler <graph_scheduler.scheduler.Scheduler>` at any time,
and take effect immediately for the execution of that Scheduler. Most pre-specified Conditions have one or more
arguments that must be specified to achieve the desired behavior. Many Conditions are also associated with an
`owner <Condition.owner>` attribute (a node to which the Condition belongs). Schedulers maintain the data
used to test for satisfaction of Condition, independent in different `execution contexts <default_execution_id>`. The Scheduler is generally
responsible for ensuring that Conditions have access to the necessary data.
When pre-specified Conditions are instantiated within a call to the `add_condition` method of a `Scheduler` or `ConditionSet`,
the Condition's `owner <Condition.owner>` is determined through
context and assigned automatically, as in the following example::

    my_scheduler.add_condition(A, EveryNPasses(1))
    my_scheduler.add_condition(B, EveryNCalls(A, 2))
    my_scheduler.add_condition(C, EveryNCalls(B, 2))

Here, ``EveryNCalls(A, 2)`` for example, is assigned the `owner` ``B``.

.. _Condition_Custom:

*Custom Conditions*
~~~~~~~~~~~~~~~~~~~

Custom Conditions can be created by calling the constructor for the base class (`Condition()`) or one of the
`generic classes <Conditions_Generic>`,  and assigning a function to the **func** argument and any arguments it
requires to the **args** and/or **kwargs** arguments (for formal or keyword arguments, respectively). The function
is called with **args** and **kwargs** by the `Scheduler` on each `PASS` through its `consideration_queue`, and the result is
used to determine whether the associated node is allowed to execute on that `PASS`. Custom Conditions allow
arbitrary schedules to be created, in which the execution of each node can depend on one or more attributes of
any other node in the graph.

.. _Condition_Recurrent_Example:

For example, the following script fragment creates a custom Condition in which ``node_A`` is scheduled to wait to
execute until ``node_B`` has "converged" (that is, settled to the point that none of
its elements has changed in value more than a specified amount since the previous `CONSIDERATION_SET_EXECUTION`)::

    def converge(node, thresh):
        for val in node.delta:
            if abs(val) >= thresh:
                return False
        return True
    epsilon = 0.01
    my_scheduler.add_condition(node_A, NWhen(Condition(converge, node_B, epsilon), 1))

In the example, a function ``converge`` is defined that references the ``delta`` attribute of
a node (which reports the change in its ``value``). The function is assigned to
the standard :class:`Condition` with ``node_A`` and ``epsilon`` as its arguments, and `composite Condition <Conditions_Composite>`
`NWhen` (which is satisfied the first N times after its condition becomes true),  The Condition is assigned to ``node_B``,
thus scheduling it to execute one time when all of the elements of ``node_A`` have changed by less than ``epsilon``.

.. _Condition_Structure:

Structure
---------

The `Scheduler <graph_scheduler.scheduler.Scheduler>` associates every node with a Condition.  If a node has not been explicitly assigned a
Condition, it is assigned a Condition that causes it to be executed whenever it is `under consideration <Scheduler_Algorithm>`
and all its structural parents have been executed at least once since the node's last execution.
Condition subclasses (`listed below <Condition_Pre-Specified_List>`)
provide a standard set of Conditions that can be implemented simply by specifying their parameter(s). There are
six types:

  * `Generic <Conditions_Generic>` - satisfied when a `user-specified function and set of arguments <Condition_Custom>`
    evaluates to `True`;
  * `Static <Conditions_Static>` - satisfied either always or never;
  * `Composite <Conditions_Composite>` - satisfied based on one or more other Conditions;
  * `Time-based <Conditions_Time_Based>` - satisfied based on the current count of units of time at a specified
    `TimeScale`;
  * `Node-based <Conditions_Node_Based>` - based on the execution or state of other nodes.
  * `Convenience <Conditions_Convenience>` - based on other Conditions, condensed for convenience

.. _Condition_Pre-Specified_List:

*List of Pre-specified Conditions*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. note::
    The optional `TimeScale` argument in many `Conditions <Condition>` specifies the unit of time over which the
    Condition operates;  the default value is `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` for all Conditions except those with "EnvironmentStateUpdate"
    in their name, for which it is `ENVIRONMENT_SEQUENCE`.


.. _Conditions_Generic:

**Generic Conditions** (used to construct `custom Conditions <Condition_Custom>`):

    * `While` (func, *args, **kwargs)
      satisfied whenever the specified function (or callable) called with args and/or kwargs evaluates to :keyword:`True`.
      Equivalent to ``Condition(func, *args, **kwargs)``

    * `WhileNot` (func, *args, **kwargs)
      satisfied whenever the specified function (or callable) called with args and/or kwargs evaluates to :keyword:`False`.
      Equivalent to ``Not(Condition(func, *args, **kwargs))``

.. _Conditions_Static:

**Static Conditions** (independent of other Conditions, nodes or time):

    * `Always`
      always satisfied.

    * `Never`
      never satisfied.


.. _Conditions_Composite:

**Composite Conditions** (based on one or more other Conditions):

    * `All` (*Conditions)
      satisfied whenever all of the specified Conditions are satisfied.

    * `Any` (*Conditions)
      satisfied whenever any of the specified Conditions are satisfied.

    * `Not` (Condition)
      satisfied whenever the specified Condition is not satisfied.

    * `NWhen` (Condition, int)
      satisfied the first specified number of times the specified Condition is satisfied.


.. _Conditions_Time_Based:

**Time-Based Conditions** (based on the count of units of time at a
specified `TimeScale` or `Time <Scheduler_Absolute_Time>`):

    * `TimeInterval` ([`pint.Quantity`, `pint.Quantity`, `pint.Quantity`])
      satisfied every time an optional amount of absolute time has
      passed in between an optional specified range

    * `TimeTermination` (`pint.Quantity`)
      satisfied after the given absolute time

    * `BeforeConsiderationSetExecution` (int[, TimeScale])
      satisfied any time before the specified `CONSIDERATION_SET_EXECUTION` occurs.

    * `AtConsiderationSetExecution` (int[, TimeScale])
      satisfied only during the specified `CONSIDERATION_SET_EXECUTION`.

    * `AfterConsiderationSetExecution` (int[, TimeScale])
      satisfied any time after the specified `CONSIDERATION_SET_EXECUTION` has occurred.

    * `AfterNConsiderationSetExecutions` (int[, TimeScale])
      satisfied when or any time after the specified number of `CONSIDERATION_SET_EXECUTION`\\ s has occurred.

    * `BeforePass` (int[, TimeScale])
      satisfied any time before the specified `PASS` occurs.

    * `AtPass` (int[, TimeScale])
      satisfied only during the specified `PASS`.

    * `AfterPass` (int[, TimeScale])
      satisfied any time after the specified `PASS` has occurred.

    * `AfterNPasses` (int[, TimeScale])
      satisfied when or any time after the specified number of `PASS`\\ es has occurred.

    * `EveryNPasses` (int[, TimeScale])
      satisfied every time the specified number of `PASS`\\ es occurs.

    * `BeforeEnvironmentStateUpdate` (int[, TimeScale])
      satisfied any time before the specified `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` occurs.

    * `AtEnvironmentStateUpdate` (int[, TimeScale])
      satisfied any time during the specified `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`.

    * `AfterEnvironmentStateUpdate` (int[, TimeScale])
      satisfied any time after the specified `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` occurs.

    * `AfterNEnvironmentStateUpdates` (int[, TimeScale])
      satisfied any time after the specified number of `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s has occurred.

    * `AtEnvironmentSequence` (int)
      satisfied any time during the specified `ENVIRONMENT_SEQUENCE`.

    * `AfterEnvironmentSequence` (int)
      satisfied any time after the specified `ENVIRONMENT_SEQUENCE` occurs.

    * `AfterNEnvironmentSequences` (int)
      satisfied any time after the specified number of `ENVIRONMENT_SEQUENCE`\\ s has occurred.

.. _Conditions_Node_Based:

**Node-Based Conditions** (based on the execution or state of other nodes):


    * `BeforeNCalls` (node, int[, TimeScale])
      satisfied any time before the specified node has executed the specified number of times.

    * `AtNCalls` (node, int[, TimeScale])
      satisfied when the specified node has executed the specified number of times.

    * `AfterCall` (node, int[, TimeScale])
      satisfied any time after the node has executed the specified number of times.

    * `AfterNCalls` (node, int[, TimeScale])
      satisfied when or any time after the node has executed the specified number of times.

    * `AfterNCallsCombined` (*nodes, int[, TimeScale])
      satisfied when or any time after the specified nodes have executed the specified number
      of times among themselves, in total.

    * `EveryNCalls` (node, int[, TimeScale])
      satisfied when the specified node has executed the specified number of times since the
      last time `owner` has run.

    * `JustRan` (node)
      satisfied if the specified node was assigned to run in the previous `CONSIDERATION_SET_EXECUTION`.

    * `AllHaveRun` (*nodes)
      satisfied when all of the specified nodes have executed at least once.

    * `WhenFinished` (node)
      satisfied when the specified node has set its `is_finished` attribute to `True`.

    * `WhenFinishedAny` (*nodes)
      satisfied when any of the specified nodes has set their `is_finished` attribute to `True`.

    * `WhenFinishedAll` (*nodes)
      satisfied when all of the specified nodes have set their `is_finished` attributes to `True`.

.. _Conditions_Convenience:

**Convenience Conditions** (based on other Conditions, condensed for convenience)


    * `AtEnvironmentStateUpdateStart`
      satisfied at the beginning of an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` (`AtPass(0) <AtPass>`)

    * `AtEnvironmentStateUpdateNStart`
      satisfied on `PASS` 0 of the specified `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` counted using 'TimeScale`

    * `AtEnvironmentSequenceStart`
      satisfied at the beginning of an `ENVIRONMENT_SEQUENCE`

    * `AtEnvironmentSequenceNStart`
      satisfied on `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 0 of the specified `ENVIRONMENT_SEQUENCE` counted using 'TimeScale`


.. Condition_Execution:

Execution
---------

When the `Scheduler` `runs <Schedule_Execution>`, it makes a sequential `PASS` through its `consideration_queue`,
evaluating each `consideration_set <consideration_set>` in the queue to determine which nodes should be assigned
to execute. It evaluates the nodes in each set by calling the `is_satisfied` method of the Condition associated
with each of those nodes.  If it returns `True`, then the node is assigned to the execution set for the
`CONSIDERATION_SET_EXECUTION` of execution generated by that `PASS`.  Otherwise, the node is not executed.

.. _Condition_Class_Reference:

Class Reference
---------------

"""

import collections
import itertools
import logging
import typing
import warnings

import pint

from graph_scheduler import _unit_registry
from graph_scheduler.time import TimeScale
from graph_scheduler.utilities import call_with_pruned_args

__all__ = [
    'AfterCall', 'AfterNCalls', 'AfterNCallsCombined', 'AfterNPasses', 'AfterNConsiderationSetExecutions', 'AfterNEnvironmentStateUpdates', 'AfterPass',
    'AtEnvironmentSequence', 'AfterEnvironmentSequence', 'AfterNEnvironmentSequences', 'AfterConsiderationSetExecution', 'AfterEnvironmentStateUpdate', 'All', 'AllHaveRun', 'Always', 'And', 'Any',
    'AtNCalls','AtPass', 'AtEnvironmentSequenceStart', 'AtEnvironmentSequenceNStart', 'AtConsiderationSetExecution', 'AtEnvironmentStateUpdate',
    'AtEnvironmentStateUpdateStart', 'AtEnvironmentStateUpdateNStart', 'BeforeNCalls', 'BeforePass', 'BeforeConsiderationSetExecution', 'BeforeEnvironmentStateUpdate',
    'Condition','ConditionError', 'ConditionSet', 'EveryNCalls', 'EveryNPasses',
    'JustRan', 'Never', 'Not', 'NWhen', 'Or', 'WhenFinished', 'WhenFinishedAll', 'WhenFinishedAny', 'While', 'WhileNot', 'TimeInterval', 'TimeTermination'
]

logger = logging.getLogger(__name__)


_pint_all_time_units = sorted(
    set([
        getattr(_unit_registry, f'{x}s')
        for x in _unit_registry._prefixes.keys()
    ]),
    reverse=True
)


def _quantity_as_integer(q):
    rounded = round(q.m, _unit_registry.precision)
    if rounded == int(rounded):
        return int(rounded) * q.u
    else:
        raise ValueError(f'{q} cannot be safely converted to integer magnitude')


def _reduce_quantity_to_integer(q):
    # find the largest time unit for which q can be expressed as an integer
    for u in filter(lambda u: u <= q.u, _pint_all_time_units):
        try:
            return _quantity_as_integer(q.to(u))
        except ValueError:
            pass

    return q


def _parse_absolute_unit(n, unit=None):
    def _get_quantity(n, unit):
        if isinstance(n, _unit_registry.Quantity):
            return n

        if isinstance(n, str):
            try:
                full_quantity = _unit_registry.Quantity(n)
            except pint.errors.UndefinedUnitError:
                pass
            else:
                # n is an actual full quantity (e.g. '1ms') not just a number ('1')
                if full_quantity.u != _unit_registry.Unit('dimensionless'):
                    return full_quantity
                else:
                    try:
                        n = int(n)
                    except ValueError:
                        n = float(n)

        try:
            # handle string representation of a unit
            unit = getattr(_unit_registry, unit)
        except TypeError:
            pass

        assert isinstance(unit, _unit_registry.Unit)
        n = n * unit

        return n

    if n is None:
        return n

    # store as the base integer quantity
    return _reduce_quantity_to_integer(_get_quantity(n, unit))


class ConditionError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class ConditionSet(object):
    """Used in conjunction with a `Scheduler <graph_scheduler.scheduler.Scheduler>` to store the `Conditions <Condition>` associated with a node.

    Arguments
    ---------

    conditions : Dict[node: `Condition`]
        specifies an iterable collection of nodes and the `Conditions <Condition>` associated
        with each.

    Attributes
    ----------

    conditions : Dict[node: `Condition`]
        the key of each entry is a node, and its value is the `Condition <Condition>` associated
        with that node.  Conditions can be added to the
        ConditionSet using the ConditionSet's `add_condition` method.

    """
    def __init__(self, conditions=None):
        self.conditions = {}

        if conditions is not None:
            self.add_condition_set(conditions)

    def __contains__(self, item):
        return item in self.conditions

    def __repr__(self):
        condition_str = '\n\t'.join([f'{owner}: {condition}' for owner, condition in self.conditions.items()])
        return '{0}({1}{2}{3})'.format(
            self.__class__.__name__,
            '\n\t' if len(condition_str) > 0 else '',
            condition_str,
            '\n' if len(condition_str) > 0 else ''
        )

    def __iter__(self):
        return iter(self.conditions)

    def __getitem__(self, key):
        return self.conditions[key]

    def __setitem__(self, key, value):
        self.conditions[key] = value

    def add_condition(self, owner, condition):
        """
        Adds a `Condition` to the ConditionSet. If **owner** already has a Condition, it is overwritten
        with the new one. If you want to add multiple conditions to a single owner, use the
        `composite Conditions <Conditions_Composite>` to accurately specify the desired behavior.

        Arguments
        ---------

        owner : node
            specifies the node with which the **condition** should be associated. **condition**
            will govern the execution behavior of **owner**

        condition : Condition
            specifies the Condition, associated with the **owner** to be added to the ConditionSet.
        """
        condition.owner = owner
        self.conditions[owner] = condition

    def add_condition_set(self, conditions):
        """
        Adds a set of `Conditions <Condition>` (in the form of a dict or another ConditionSet) to the ConditionSet.
        Any Condition added here will overwrite an existing Condition for a given owner.
        If you want to add multiple conditions to a single owner, add a single `Composite Condition <Conditions_Composite>`
        to accurately specify the desired behavior.

        Arguments
        ---------

        conditions : dict[node: `Condition`], `ConditionSet`
            specifies collection of Conditions to be added to this ConditionSet,

            if a dict is provided:
                each entry should map an owner node (the node whose execution behavior will be
                governed) to a `Condition <Condition>`

        """
        for owner in conditions:
            self.add_condition(owner, conditions[owner])


class Condition:
    """
    Used in conjunction with a :class:`Scheduler` to specify the condition under which a node should be
    allowed to execute.

    Arguments
    ---------

    func : callable
        specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

    args : *args
        specifies formal arguments to pass to `func` when the Condition is evaluated.

    kwargs : **kwargs
        specifies keyword arguments to pass to `func` when the Condition is evaluated.

    Attributes
    ----------

    owner (node):
        the node with which the Condition is associated, and the execution of which it determines.

    """
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

        self._owner = None

        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __str__(self):
        return '{0}({1}{2})'.format(
            self.__class__.__name__,
            ', '.join([str(arg) for arg in self.args]) if len(self.args) > 0 else '',
            ', {0}'.format(self.kwargs) if len(self.kwargs) > 0 else ''
        )

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        logger.debug('Condition ({0}) setting owner to {1}'.format(type(self).__name__, value))
        self._owner = value

    def is_satisfied(self, *args, execution_id=None, **kwargs):
        """
        the function called to determine satisfaction of this Condition.

        Arguments
        ---------
        args : *args
            specifies additional formal arguments to pass to `func` when the Condition is evaluated.
            these are appended to the **args** specified at instantiation of this Condition

        kwargs : **kwargs
            specifies additional keyword arguments to pass to `func` when the Condition is evaluated.
            these are added to the **kwargs** specified at instantiation of this Condition

        Returns
        -------
            True - if the Condition is satisfied
            False - if the Condition is not satisfied
        """
        # update so that kwargs can override self.kwargs
        kwargs_to_pass = self.kwargs.copy()
        kwargs_to_pass.update(kwargs)

        return call_with_pruned_args(
            self.func,
            *self.args,
            *args,
            execution_id=execution_id,
            **kwargs_to_pass
        )

    @property
    def absolute_intervals(self):
        """
        In absolute time, repeated intervals for satisfaction or
        unsatisfaction of this Condition or those it contains
        """
        return []

    @property
    def absolute_fixed_points(self):
        """
        In absolute time, specific time points for satisfaction or
        unsatisfaction of this Condition or those it contains
        """
        return []

    @property
    def is_absolute(self):
        return False


class AbsoluteCondition(Condition):
    @property
    def is_absolute(self):
        return True


class _DependencyValidation:
    @Condition.owner.setter
    def owner(self, value):
        try:
            # "dependency" or "dependencies" is always the first positional argument
            if not isinstance(self.args[0], collections.abc.Iterable):
                dependencies = [self.args[0]]
            else:
                dependencies = self.args[0]
        except IndexError:
            pass
        else:
            if value in dependencies:
                warnings.warn(
                    f'{self} is dependent on {value}, but you are assigning {value} as its owner.'
                    ' This may result in infinite loops or unknown behavior.',
                    stacklevel=5
                )

        self._owner = value


#########################################################################################################
# Included Conditions
#########################################################################################################

######################################################################
# Generic Conditions
#   - convenience wrappers
######################################################################


While = Condition


class WhileNot(Condition):
    """
    WhileNot

    Parameters:

        func : callable
            specifies function to be called when the Condition is evaluated, to determine whether it is currently satisfied.

        args : *args
            specifies formal arguments to pass to `func` when the Condition is evaluated.

        kwargs : **kwargs
            specifies keyword arguments to pass to `func` when the Condition is evaluated.

    Satisfied when:

        - **func** is False

    """
    def __init__(self, func, *args, **kwargs):
        def inner_func(*args, **kwargs):
            return not call_with_pruned_args(func, *args, **kwargs)
        super().__init__(inner_func, *args, **kwargs)

######################################################################
# Static Conditions
#   - independent of nodes and time
######################################################################


class Always(Condition):
    """Always

    Parameters:

        none

    Satisfied when:

        - always satisfied.

    """
    def __init__(self):
        super().__init__(lambda: True)


class Never(Condition):
    """Never

    Parameters:

        none

    Satisfied when:

        - never satisfied.
    """
    def __init__(self):
        super().__init__(lambda: False)

######################################################################
# Composite Conditions
#   - based on other Conditions
######################################################################


class CompositeCondition(Condition):
    @Condition.owner.setter
    def owner(self, value):
        super(CompositeCondition, CompositeCondition).owner.__set__(self, value)
        for cond in self.args:
            logger.debug('owner setter: Setting owner of {0} to ({1})'.format(cond, value))
            if cond.owner is None:
                cond.owner = value

    @property
    def absolute_intervals(self):
        return list(itertools.chain.from_iterable([
            c.absolute_intervals
            for c in filter(lambda a: a.is_absolute, self.args)
        ]))

    @property
    def absolute_fixed_points(self):
        return list(itertools.chain.from_iterable([
            c.absolute_fixed_points
            for c in filter(lambda a: a.is_absolute, self.args)
        ]))

    @property
    def is_absolute(self):
        return any(a.is_absolute for a in self.args)


class All(CompositeCondition):
    """All

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - all of the Conditions in args are satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = All(*conditions)

    """
    def __init__(self, *args, **dependencies):
        args += tuple(*[v for k, v in dependencies.items()])
        super().__init__(self.satis, *args)

    def satis(self, *conds, **kwargs):
        for cond in conds:
            if not cond.is_satisfied(**kwargs):
                return False
        return True


class Any(CompositeCondition):
    """Any

    Parameters:

        args: one or more `Conditions <Condition>`

    Satisfied when:

        - one or more of the Conditions in **args** is satisfied.

    Notes:

        - To initialize with a list (for example)::

            conditions = [AfterNCalls(node, 5) for node in node_list]

          unpack the list to supply its members as args::

           composite_condition = Any(*conditions)

    """
    def __init__(self, *args, **dependencies):
        args += tuple(*[v for k, v in dependencies.items()])
        super().__init__(self.satis, *args)

    def satis(self, *conds, **kwargs):
        for cond in conds:
            if cond.is_satisfied(**kwargs):
                return True
        return False


And = All
Or = Any


class Not(Condition):
    """Not

    Parameters:

        condition(Condition): a `Condition`

    Satisfied when:

        - **condition** is not satisfied.

    """
    def __init__(self, condition):
        self.condition = condition

        def inner_func(condition, *args, **kwargs):
            return not condition.is_satisfied(*args, **kwargs)
        super().__init__(inner_func, condition)

    @Condition.owner.setter
    def owner(self, value):
        super(Not, Not).owner.__set__(self, value)
        self.condition.owner = value


class NWhen(Condition):
    """NWhen

    Parameters:

        condition(Condition): a `Condition`

        n(int): the maximum number of times this condition will be satisfied

    Satisfied when:

        - the first **n** times **condition** is satisfied upon evaluation

    """
    def __init__(self, condition, n=1):
        self.satisfactions = {}
        self.condition = condition

        super().__init__(self.satis, condition, n)

    @Condition.owner.setter
    def owner(self, value):
        super(NWhen, NWhen).owner.__set__(self, value)
        self.condition.owner = value

    def satis(self, condition, n, *args, scheduler=None, execution_id=None, **kwargs):
        if execution_id not in self.satisfactions:
            self.satisfactions[execution_id] = 0

        if self.satisfactions[execution_id] < n:
            if call_with_pruned_args(condition.is_satisfied, *args, scheduler=scheduler, execution_id=execution_id, **kwargs):
                self.satisfactions[execution_id] += 1
                return True
        return False


######################################################################
# Time-based Conditions
#   - satisfied based only on time
######################################################################

class TimeInterval(AbsoluteCondition):
    """TimeInterval

    Attributes:

        repeat
            the interval between *unit*s where this condition can be
            satisfied

        start
            the time at/after which this condition can be
            satisfied

        end
            the time at/fter which this condition can be
            satisfied

        unit
            the `pint.Unit` to use for scalar values of *repeat*,
            *start*, and *end*

        start_inclusive
            if True, *start* allows satisfaction exactly at the time
            corresponding to *start*. if False, satisfaction can occur
            only after *start*

        end_inclusive
            if True, *end* allows satisfaction exactly until the time
            corresponding to *end*. if False, satisfaction can occur
            only before *end*


    Satisfied when:

        Every *repeat* units of time at/after *start* and before/through
        *end*

    Notes:

        Using a `TimeInterval` as a
        `termination Condition <Scheduler_Termination_Conditions>` may
        result in unexpected behavior. The user may be inclined to
        create **TimeInterval(end=x)** to terminate at time **x**, but
        this will do the opposite and be true only and always until time
        **x**, terminating at any time before **x**. If in doubt, use
        `TimeTermination` instead.

        If the scheduler is not set to `exact_time_mode = True`,
        *start_inclusive* and *end_inclusive* may not behave as
        expected. See `Scheduler_Exact_Time` for more info.
    """
    def __init__(
        self,
        repeat: typing.Union[int, str, pint.Quantity] = None,
        start: typing.Union[int, str, pint.Quantity] = None,
        end: typing.Union[int, str, pint.Quantity] = None,
        unit: typing.Union[str, pint.Unit] = _unit_registry.ms,
        start_inclusive: bool = True,
        end_inclusive: bool = True
    ):
        if repeat is start is end is None:
            raise ConditionError(
                'At least one of "repeat", "start", "end" must be specified'
            )

        if start is not None and end is not None and start > end:
            raise ConditionError(f'Start ({start}) is later than {end}')

        repeat = _parse_absolute_unit(repeat, unit)
        start = _parse_absolute_unit(start, unit)
        end = _parse_absolute_unit(end, unit)

        def func(scheduler, execution_id):
            satisfied = True
            clock = scheduler.get_clock(execution_id)

            if scheduler._in_exact_time_mode and start is not None:
                offset = start
            else:
                for i, cs in enumerate(scheduler.consideration_queue):
                    if self.owner in cs:
                        offset = i * clock.time.absolute_interval

            if repeat is not None:
                satisfied &= round(
                    (clock.time.absolute - offset) % repeat,
                    _unit_registry.precision
                ) == 0

            if start is not None:
                if start_inclusive:
                    satisfied &= clock.time.absolute >= start
                else:
                    satisfied &= clock.time.absolute > start

            if end is not None:
                if end_inclusive:
                    satisfied &= clock.time.absolute <= end
                else:
                    satisfied &= clock.time.absolute < end

            return satisfied

        super().__init__(
            func,
            repeat=repeat,
            start=start,
            end=end,
            unit=unit,
            start_inclusive=start_inclusive,
            end_inclusive=end_inclusive
        )

    @property
    def absolute_intervals(self):
        return [self.repeat] if self.repeat is not None else []

    @property
    def absolute_fixed_points(self):
        fp = []
        if self.start_inclusive and self.start is not None:
            fp.append(self.start)
        if self.end_inclusive and self.end is not None:
            fp.append(self.end)

        return fp


class TimeTermination(AbsoluteCondition):
    """TimeTermination

    Attributes:

        t
            the time at/after which this condition is satisfied

        unit
            the `pint.Unit` to use for scalar values of *t*, *start*,
            and *end*

        start_inclusive
            if True, the condition is satisfied exactly at the time
            corresponding to *t*. if False, satisfaction can occur
            only after *t*

    Satisfied when:

        At/After time *t*
    """
    def __init__(
        self,
        t: typing.Union[int, str, pint.Quantity],
        inclusive: bool = True,
        unit: typing.Union[str, pint.Unit] = _unit_registry.ms,
    ):
        t = _parse_absolute_unit(t, unit)

        def func(scheduler, execution_id):
            clock = scheduler.get_clock(execution_id)

            if inclusive:
                return clock.time.absolute >= t
            else:
                return clock.time.absolute > t

        super().__init__(func, t=t, inclusive=inclusive)

    @property
    def absolute_fixed_points(self):
        return [self.t] if self.inclusive else []


class BeforeConsiderationSetExecution(Condition):
    """BeforeConsiderationSetExecution

    Parameters:

        n(int): the 'CONSIDERATION_SET_EXECUTION' before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `CONSIDERATION_SET_EXECUTION`\\ s (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - at most n-1 `CONSIDERATION_SET_EXECUTION`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `CONSIDERATION_SET_EXECUTION` is 0, the second `CONSIDERATION_SET_EXECUTION` is 1, etc.);
          so, `BeforeConsiderationSetExecution(2)` is satisfied at `CONSIDERATION_SET_EXECUTION` 0 and `CONSIDERATION_SET_EXECUTION` 1.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.CONSIDERATION_SET_EXECUTION, time_scale) < n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class AtConsiderationSetExecution(Condition):
    """AtConsiderationSetExecution

    Parameters:

        n(int): the `CONSIDERATION_SET_EXECUTION` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `CONSIDERATION_SET_EXECUTION`\\ s (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - exactly n `CONSIDERATION_SET_EXECUTION`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first 'CONSIDERATION_SET_EXECUTION' is pass 0, the second 'CONSIDERATION_SET_EXECUTION' is 1, etc.);
          so, `AtConsiderationSetExecution(1)` is satisfied when a single `CONSIDERATION_SET_EXECUTION` (`CONSIDERATION_SET_EXECUTION` 0) has occurred, and `AtConsiderationSetExecution(2)` is satisfied
          when two `CONSIDERATION_SET_EXECUTION`\\ s have occurred (`CONSIDERATION_SET_EXECUTION` 0 and `CONSIDERATION_SET_EXECUTION` 1), etc..

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.CONSIDERATION_SET_EXECUTION, time_scale) == n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterConsiderationSetExecution(Condition):
    """AfterConsiderationSetExecution

    Parameters:

        n(int): the `CONSIDERATION_SET_EXECUTION` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `CONSIDERATION_SET_EXECUTION`\\ s (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - at least n+1 `CONSIDERATION_SET_EXECUTION`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScals are zero-indexed (that is, the first `CONSIDERATION_SET_EXECUTION` is 0, the second `CONSIDERATION_SET_EXECUTION` is 1, etc.); so,
          `AfterConsiderationSetExecution(1)` is satisfied after `CONSIDERATION_SET_EXECUTION` 1 has occurred and thereafter (i.e., in `CONSIDERATION_SET_EXECUTION`\\ s 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.CONSIDERATION_SET_EXECUTION, time_scale) > n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class AfterNConsiderationSetExecutions(Condition):
    """AfterNConsiderationSetExecutions

    Parameters:

        n(int): the number of `CONSIDERATION_SET_EXECUTION`\\ s after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `CONSIDERATION_SET_EXECUTION`\\ s (default: TimeScale.ENVIRONMENT_STATE_UPDATE)


    Satisfied when:

        - at least n `CONSIDERATION_SET_EXECUTION`\\ s have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.CONSIDERATION_SET_EXECUTION, time_scale) >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class BeforePass(Condition):
    """BeforePass

    Parameters:

        n(int): the 'PASS' before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - at most n-1 `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `PASS` is 0, the second `PASS` is 1, etc.);
          so, `BeforePass(2)` is satisfied at `PASS` 0 and `PASS` 1.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.PASS, time_scale) < n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class AtPass(Condition):
    """AtPass

    Parameters:

        n(int): the `PASS` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - exactly n `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first 'PASS' is pass 0, the second 'PASS' is 1, etc.);
          so, `AtPass(1)` is satisfied when a single `PASS` (`PASS` 0) has occurred, and `AtPass(2)` is satisfied
          when two `PASS`\\ es have occurred (`PASS` 0 and `PASS` 1), etc..

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.PASS, time_scale) == n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterPass(Condition):
    """AfterPass

    Parameters:

        n(int): the `PASS` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - at least n+1 `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `PASS` is 0, the second `PASS` is 1, etc.); so,
          `AfterPass(1)` is satisfied after `PASS` 1 has occurred and thereafter (i.e., in `PASS`\\ es 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.PASS, time_scale) > n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class AfterNPasses(Condition):
    """AfterNPasses

    Parameters:

        n(int): the number of `PASS`\\ es after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.ENVIRONMENT_STATE_UPDATE)


    Satisfied when:

        - at least n `PASS`\\ es have occurred within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.PASS, time_scale) >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class EveryNPasses(Condition):
    """EveryNPasses

    Parameters:

        n(int): the frequency of passes with which this condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `PASS`\\ es (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - `PASS` 0

        - the specified number of `PASS`\\ es that has occurred within a unit of time (at the `TimeScale` specified by
          **time_scale**) is evenly divisible by n.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.PASS, time_scale) % n == 0
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class BeforeEnvironmentStateUpdate(Condition):
    """BeforeEnvironmentStateUpdate

    Parameters:

        n(int): the `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s
        (default: TimeScale.ENVIRONMENT_SEQUENCE)

    Satisfied when:

        - at most n-1 `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 0, the second
          `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 1, etc.); so, `BeforeEnvironmentStateUpdate(2)` is satisfied at `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 0
          and `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 1.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_SEQUENCE):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.ENVIRONMENT_STATE_UPDATE, time_scale) < n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AtEnvironmentStateUpdate(Condition):
    """AtEnvironmentStateUpdate

    Parameters:

        n(int): the `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s
        (default: TimeScale.ENVIRONMENT_SEQUENCE)

    Satisfied when:

        - exactly n `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 0,
          the second `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 1, etc.); so, `AtEnvironmentStateUpdate(1)` is satisfied when one
          `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` (`ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 0) has already occurred.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_SEQUENCE):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.ENVIRONMENT_STATE_UPDATE, time_scale) == n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterEnvironmentStateUpdate(Condition):
    """AfterEnvironmentStateUpdate

    Parameters:

        n(int): the `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s.
        (default: TimeScale.ENVIRONMENT_SEQUENCE)

    Satisfied when:

        - at least n+1 `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s have occurred within one unit of time at the `TimeScale`
          specified by **time_scale**.

    Notes:

        - Counts of TimeScales are zero-indexed (that is, the first `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 0, the second
        `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is 1, etc.); so,  `AfterPass(1)` is satisfied after `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 1
        has occurred and thereafter (i.e., in `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s 2, 3, 4, etc.).

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_SEQUENCE):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.ENVIRONMENT_STATE_UPDATE, time_scale) > n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterNEnvironmentStateUpdates(Condition):
    """AfterNEnvironmentStateUpdates

    Parameters:

        n(int): the number of `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s
        (default: TimeScale.ENVIRONMENT_SEQUENCE)

    Satisfied when:

        - at least n `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s have occured within one unit of time at the `TimeScale`
          specified by **time_scale**.

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_SEQUENCE):
        def func(n, time_scale, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).get_total_times_relative(TimeScale.ENVIRONMENT_STATE_UPDATE, time_scale) >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n, time_scale)


class AtEnvironmentSequence(Condition):
    """AtEnvironmentSequence

    Parameters:

        n(int): the `ENVIRONMENT_SEQUENCE` at which the Condition is satisfied

    Satisfied when:

        - exactly n `ENVIRONMENT_SEQUENCE`\\ s have occurred.

    """
    def __init__(self, n):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).time.environment_sequence == n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterEnvironmentSequence(Condition):
    """AfterEnvironmentSequence

    Parameters:

        n(int): the `ENVIRONMENT_SEQUENCE` after which the Condition is satisfied

    Satisfied when:

        - at least n+1 `ENVIRONMENT_SEQUENCE`\\ s have occurred.

    """
    def __init__(self, n):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).time.environment_sequence > n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)


class AfterNEnvironmentSequences(Condition):
    """AfterNEnvironmentStateUpdates

    Parameters:

        n(int): the number of `ENVIRONMENT_SEQUENCE`\\ s after which the Condition is satisfied

    Satisfied when:

        - at least n `ENVIRONMENT_SEQUENCE`\\ s have occured.

    """

    def __init__(self, n):
        def func(n, scheduler=None, execution_id=None):
            try:
                return scheduler.get_clock(execution_id).time.environment_sequence >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, n)

######################################################################
# Node-based Conditions
#   - satisfied based on executions or state of nodes
######################################################################


class BeforeNCalls(_DependencyValidation, Condition):
    """BeforeNCalls

    Parameters:

        dependency(node):  the node on which the Condition depends

        n(int): the number of executions of **dependency** before which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - the node specified in **dependency** has executed at most n-1 times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        self.time_scale = time_scale

        def func(dependency, n, scheduler=None, execution_id=None):
            try:
                num_calls = scheduler.counts_total[execution_id][time_scale][dependency]
                logger.debug('{0} has reached {1} num_calls in {2}'.format(dependency, num_calls, time_scale.name))
                return num_calls < n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, dependency, n)

# NOTE:
# The behavior of AtNCalls is not desired (i.e. depending on the order nodes are checked, B running AtNCalls(A, x))
# may run on both the xth and x+1st call of A; if A and B are not parent-child
# A fix could invalidate key assumptions and affect many other conditions
# Since this condition is unlikely to be used, it's best to leave it for now


class AtNCalls(_DependencyValidation, Condition):
    """AtNCalls

    Parameters:

        dependency(node):  the node on which the Condition depends

        n(int): the number of executions of **dependency** at which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - the node specified in **dependency** has executed exactly n times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        self.time_scale = time_scale

        def func(dependency, n, scheduler=None, execution_id=None):
            try:
                num_calls = scheduler.counts_total[execution_id][time_scale][dependency]
                logger.debug('{0} has reached {1} num_calls in {2}'.format(dependency, num_calls, time_scale.name))
                return num_calls == n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, dependency, n)


class AfterCall(_DependencyValidation, Condition):
    """AfterCall

    Parameters:

        dependency(node):  the node on which the Condition depends

        n(int): the number of executions of **dependency** after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - the node specified in **dependency** has executed at least n+1 times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        def func(dependency, n, scheduler=None, execution_id=None):
            try:
                num_calls = scheduler.counts_total[execution_id][time_scale][dependency]
                logger.debug('{0} has reached {1} num_calls in {2}'.format(dependency, num_calls, time_scale.name))
                return num_calls > n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, dependency, n)


class AfterNCalls(_DependencyValidation, Condition):
    """AfterNCalls

    Parameters:

        dependency(node):  the node on which the Condition depends

        n(int): the number of executions of **dependency** after which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - the node specified in **dependency** has executed at least n times
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, dependency, n, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        self.time_scale = time_scale

        def func(dependency, n, scheduler=None, execution_id=None):
            try:
                num_calls = scheduler.counts_total[execution_id][time_scale][dependency]
                logger.debug('{0} has reached {1} num_calls in {2}'.format(dependency, num_calls, time_scale.name))
                return num_calls >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, dependency, n)


class AfterNCallsCombined(_DependencyValidation, Condition):
    """AfterNCallsCombined

    Parameters:

        *nodes(nodes):  one or more nodes on which the Condition depends

        n(int): the number of combined executions of all nodes specified in **dependencies** after which the
        Condition is satisfied (default: None)

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)


    Satisfied when:

        - there have been at least n+1 executions among all of the nodes specified in **dependencies**
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, *dependencies, n=None, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        logger.debug('{0} args: deps {1}, n {2}, ts {3}'.format(type(self).__name__, dependencies, n, time_scale))

        def func(*dependencies, n=None, scheduler=None, execution_id=None):
            if n is None:
                raise ConditionError(f'{type(self).__name__}: required keyword argument n is None.')
            count_sum = 0
            for d in dependencies:
                try:
                    count_sum += scheduler.counts_total[execution_id][time_scale][d]
                    logger.debug('{0} has reached {1} num_calls in {2}'.
                                 format(d, scheduler.counts_total[execution_id][time_scale][d], time_scale.name))
                except AttributeError as e:
                    raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

            return count_sum >= n
        super().__init__(func, *dependencies, n=n)


class EveryNCalls(_DependencyValidation, Condition):
    """EveryNCalls

    Parameters:

        dependency(node):  the node on which the Condition depends

        n(int): the frequency of executions of **dependency** at which the Condition is satisfied


    Satisfied when:

        - the node specified in **dependency** has executed at least n times since the last time the
          Condition's owner executed.

        COMMENT:
            JDC: IS THE FOLLOWING TRUE OF ALL OF THE ABOVE AS WELL??
            K: No, EveryNCalls is tricky in how it needs to be implemented, because it's in a sense
                tracking the relative frequency of calls between two objects. So the idea is that the scheduler
                tracks how many executions of a node are "useable" by other nodes for EveryNCalls conditions.
                So, suppose you had something like add_condition(B, All(AfterNCalls(A, 10), EveryNCalls(A, 2))). You
                would want the AAB pattern to start happening after A has run 10 times. Useable counts allows B to see
                whether A has run enough times for it to run, and then B spends its "useable executions" of A. Then,
                A must run two more times for B to run again. If you didn't reset the counts of A useable by B
                to 0 (question below) when B runs, then in the
                above case B would continue to run every pass for the next 4 passes, because it would see an additional
                8 executions of A it could spend to execute.
            JDC: IS THIS A FORM OF MODULO?  IF SO, WOULD IT BE EASIER TO EXPLAIN IN THAT FORM?
        COMMENT

    Notes:

        - scheduler's count of each other node that is "useable" by the node is reset to 0 when the
          node runs

    """
    def __init__(self, dependency, n):
        def func(dependency, n, scheduler=None, execution_id=None):
            try:
                num_calls = scheduler.counts_useable[execution_id][dependency][self.owner]
                logger.debug('{0} has reached {1} num_calls'.format(dependency, num_calls))
                return num_calls >= n
            except AttributeError as e:
                raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')

        super().__init__(func, dependency, n)


class JustRan(_DependencyValidation, Condition):
    """JustRan

    Parameters:

        dependency(node):  the node on which the Condition depends

    Satisfied when:

        - the node specified in **dependency** executed in the previous `CONSIDERATION_SET_EXECUTION`.

    Notes:

        - This Condition can transcend divisions between `TimeScales <TimeScale>`.
          For example, if A runs in the final `CONSIDERATION_SET_EXECUTION` of an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`,
          JustRan(A) is satisfied at the beginning of the next `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`.

    """
    def __init__(self, dependency):
        def func(dependency, scheduler=None, execution_id=None):
            logger.debug(f'checking if {dependency} in previous execution step set')
            try:
                return dependency in scheduler.execution_list[execution_id][-1]
            except TypeError:
                return dependency == scheduler.execution_list[execution_id][-1]
        super().__init__(func, dependency)


class AllHaveRun(_DependencyValidation, Condition):
    """AllHaveRun

    Parameters:

        *nodes(nodes):  an iterable of nodes on which the Condition depends

        time_scale(TimeScale): the TimeScale used as basis for counting executions of **dependency**
        (default: TimeScale.ENVIRONMENT_STATE_UPDATE)

    Satisfied when:

        - all of the nodes specified in **dependencies** have executed at least once
          within one unit of time at the `TimeScale` specified by **time_scale**.

    """
    def __init__(self, *dependencies, time_scale=TimeScale.ENVIRONMENT_STATE_UPDATE):
        self.time_scale = time_scale

        def func(*dependencies, scheduler=None, execution_id=None):
            if len(dependencies) == 0:
                dependencies = scheduler.nodes
            for d in dependencies:
                try:
                    if scheduler.counts_total[execution_id][time_scale][d] < 1:
                        return False
                except AttributeError as e:
                    raise ConditionError(f'{type(self).__name__}: scheduler must be supplied to is_satisfied: {e}.')
                except KeyError as e:
                    raise ConditionError(
                        f'{type(self).__name__}: execution_id ({execution_id}) must both be specified, and '
                        f'execution_id must be in scheduler.counts_total (scheduler: {scheduler}): {e}.')
            return True
        super().__init__(func, *dependencies)


class WhenFinished(_DependencyValidation, Condition):
    """WhenFinished

    Parameters:

        dependency(node):  the node on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of the node specified in **dependencies** returns `True`.

    Notes:

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

    """
    def __init__(self, dependency):
        def func(dependency, execution_id=None):
            try:
                return dependency.is_finished(execution_id)
            except AttributeError as e:
                raise ConditionError(f'WhenFinished: Unsupported dependency type: {type(dependency)}; ({e}).')

        super().__init__(func, dependency)


class WhenFinishedAny(_DependencyValidation, Condition):
    """WhenFinishedAny

    Parameters:

        *nodes(nodes):  zero or more nodes on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of any nodes specified in **dependencies** returns `True`.

    Notes:

        - This is a convenience class; WhenFinishedAny(A, B, C) is equivalent to
          Any(WhenFinished(A), WhenFinished(B), WhenFinished(C)).
          If no nodes are specified, the condition will default to checking all of scheduler's nodes.

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

    """
    def __init__(self, *dependencies):
        def func(*dependencies, scheduler=None, execution_id=None):
            if len(dependencies) == 0:
                dependencies = scheduler.nodes
            for d in dependencies:
                try:
                    if d.is_finished(execution_id):
                        return True
                except AttributeError as e:
                    raise ConditionError(f'WhenFinishedAny: Unsupported dependency type: {type(d)}; ({e}).')
            return False

        super().__init__(func, *dependencies)


class WhenFinishedAll(_DependencyValidation, Condition):
    """WhenFinishedAll

    Parameters:

        *nodes(nodes):  zero or more nodes on which the Condition depends

    Satisfied when:

        - the `is_finished` methods of all nodes specified in **dependencies** return `True`.

    Notes:

        - This is a convenience class; WhenFinishedAny(A, B, C) is equivalent to
          All(WhenFinished(A), WhenFinished(B), WhenFinished(C)).
          If no nodes are specified, the condition will default to checking all of scheduler's nodes.

        - This is a dynamic Condition: Each node is responsible for managing its finished status on its
          own, which can occur independently of the execution of other nodes.  Therefore the satisfaction of
          this Condition) can vary arbitrarily in time.

    """
    def __init__(self, *dependencies):
        def func(*dependencies, scheduler=None, execution_id=None):
            if len(dependencies) == 0:
                dependencies = scheduler.nodes
            for d in dependencies:
                try:
                    if not d.is_finished(execution_id):
                        return False
                except AttributeError as e:
                    raise ConditionError(f'WhenFinishedAll: Unsupported dependency type: {type(d)}; ({e})')
            return True

        super().__init__(func, *dependencies)


######################################################################
# Convenience Conditions
######################################################################


class AtEnvironmentStateUpdateStart(AtPass):
    """AtEnvironmentStateUpdateStart

    Satisfied when:

        - at the beginning of an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`

    Notes:

        - identical to `AtPass(0) <AtPass>`
    """
    def __init__(self):
        super().__init__(0)

    def __str__(self):
        return '{0}()'.format(self.__class__.__name__)


class AtEnvironmentStateUpdateNStart(All):
    """AtEnvironmentStateUpdateNStart

    Parameters:

        n(int): the `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` on which the Condition is satisfied

        time_scale(TimeScale): the TimeScale used as basis for counting `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>`\\ s
        (default: TimeScale.ENVIRONMENT_SEQUENCE)

    Satisfied when:

        - on `PASS` 0 of the specified `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` counted using 'TimeScale`

    Notes:

        - identical to All(AtPass(0), AtEnvironmentStateUpdate(n, time_scale))

    """
    def __init__(self, n, time_scale=TimeScale.ENVIRONMENT_SEQUENCE):
        return super().__init__(AtPass(0), AtEnvironmentStateUpdate(n, time_scale))


class AtEnvironmentSequenceStart(AtEnvironmentStateUpdate):
    """AtEnvironmentSequenceStart

    Satisfied when:

        - at the beginning of an `ENVIRONMENT_SEQUENCE`

    Notes:

        - identical to `AtEnvironmentStateUpdate(0) <AtEnvironmentStateUpdate>`
    """
    def __init__(self):
        super().__init__(0, time_scale=TimeScale.ENVIRONMENT_SEQUENCE)

    def __str__(self):
        return '{0}()'.format(self.__class__.__name__)


class AtEnvironmentSequenceNStart(All):
    """AtEnvironmentSequenceNStart

    Parameters:

        n(int): the `ENVIRONMENT_SEQUENCE` on which the Condition is satisfied

    Satisfied when:

        - on `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` 0 of the specified `ENVIRONMENT_SEQUENCE` counted using 'TimeScale`

    Notes:

        - identical to `All(AtEnvironmentStateUpdate(0), AtEnvironmentSequence(n))`

    """
    def __init__(self, n):
        return super().__init__(AtEnvironmentStateUpdate(0), AtEnvironmentSequence(n))



import doctest

import pytest
import graph_scheduler


@pytest.mark.parametrize(
    "mod",
    [
        graph_scheduler.condition,
        graph_scheduler.scheduler,
        graph_scheduler.time,
    ]
)
def test_other_docs(mod, capsys):
    fail, total = doctest.testmod(mod, optionflags=doctest.REPORT_NDIFF)
    if fail > 0:
        captured = capsys.readouterr()
        pytest.fail("{} out of {} examples failed:\n{}\n{}".format(
            fail, total, captured.err, captured.out), pytrace=False)



"""

.. _Time_Overview:

Overview
--------

:doc:`Scheduler`\\ s maintain `Clock` objects to track time. The current time in \
relation to a :doc:`Scheduler` is stored in :class:`Clock.time <Time>` \
or :class:`Clock.simple_time <SimpleTime>`

"""

import copy
import enum
import functools
import keyword
import re
import types
import typing

from graph_scheduler import _unit_registry

__all__ = [
    'Clock', 'TimeScale', 'Time', 'SimpleTime', 'TimeHistoryTree', 'TimeScaleError', 'set_time_scale_alias', 'remove_time_scale_alias'
]


_time_scale_aliases = {}
_alias_docs_warning_str = """

.. note:: This documentation was modified from the original due to environment-specific TimeScale renamings. If there is any confusion, please see the original documentation at https://www.github.com/kmantel/graph-scheduler

"""


class TimeScaleError(Exception):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


# Time scale modes
@functools.total_ordering
class TimeScale(enum.Enum):
    """Represents divisions of time used by the `Scheduler` and `Conditions <Condition>`.

    The values of TimeScale are defined as follows (in order of increasingly coarse granularity):

    Attributes
    ----------

    CONSIDERATION_SET_EXECUTION
        the nuclear unit of time, corresponding to the execution of all nodes allowed to execute
        from a single `consideration set <Consideration_Set>` of a `Scheduler <graph_scheduler.scheduler.Scheduler>`, and which are considered to have
        executed simultaneously.

    PASS
        a full iteration through all of the consideration sets in a Scheduler's <Scheduler>
        `consideration_queue`, consisting of one or more `CONSIDERATION_SET_EXECUTIONs <CONSIDERATION_SET_EXECUTION>`, over which every node
        specified to a Scheduler <Scheduler_Creation>` is considered for execution at least once.

    ENVIRONMENT_STATE_UPDATE
        an open-ended unit of time consisting of all actions that occurs within the scope of a single
        call to `run <Scheduler.run>`

    ENVIRONMENT_SEQUENCE
        the scope of a batch of one or more `ENVIRONMENT_STATE_UPDATE
        <TimeScale.ENVIRONMENT_STATE_UPDATE>`_\\s, managed by the
        environment using the Scheduler.

    LIFE
        the scope of time since the creation of an object.
    """
    CONSIDERATION_SET_EXECUTION = 0
    PASS = 1
    ENVIRONMENT_STATE_UPDATE = 2
    ENVIRONMENT_SEQUENCE = 3
    LIFE = 4

    # ordering based on enum.OrderedEnum example
    # https://docs.python.org/3/library/enum.html#orderedenum
    # https://stackoverflow.com/a/39269589/3131666
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    @classmethod
    def get_parent(cls, time_scale):
        """
        Returns
        -------
            the TimeScale one level wider in scope than time_scale : :class:`TimeScale`
        """
        return cls(time_scale.value + 1)

    @classmethod
    def get_child(cls, time_scale):
        """
        Returns
        -------
            the TimeScale one level smaller in scope than time_scale : :class:`TimeScale`
        """
        return cls(time_scale.value - 1)


class Clock:
    """
    Stores a history of :class:`TimeScale`\\ s that have occurred, and keep track of a \
    current `Time`. Used in relation to a :doc:`Scheduler`

    Attributes
    ----------
        history : `TimeHistoryTree`
            a root `TimeHistoryTree` associated with this Clock

        simple_time : `SimpleTime`
            the current time in simple format
    """
    def __init__(self):
        self.history = TimeHistoryTree()
        self.simple_time = SimpleTime(self.time)

    def __repr__(self):
        return 'Clock({0})'.format(self.time.__repr__())

    def _increment_time(self, time_scale):
        """
        Calls `self.history.increment_time <TimeHistoryTree.increment_time>`
        """
        self.history.increment_time(time_scale)

    def get_total_times_relative(self, query_time_scale, base_time_scale, base_index=None):
        """
        Convenience simplified wrapper for `TimeHistoryTree.get_total_times_relative`

        Arguments
        ---------
            query_time_scale : :class:`TimeScale`
                the unit of time whose number of ticks to be returned

            base_time_scale : :class:`TimeScale`
                the unit of time over which the number of **query_time_scale** ticks
                should be returned

            base_index : int
                the **base_index**\\ th **base_time_scale** over which the number of
                **query_time_scale** ticks should be returned
        Returns
        -------
            the number of query_time_scale s that have occurred during the scope \
            of the base_index'th base_time_scale : int
        """
        if base_index is None:
            base_index = self.get_time_by_time_scale(base_time_scale)

        return self.history.get_total_times_relative(
            query_time_scale,
            {base_time_scale: base_index}
        )

    def get_time_by_time_scale(self, time_scale):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Returns
        -------
            the current value of the time unit corresponding to time_scale \
            for this Clock : int
        """
        return self.time._get_by_time_scale(time_scale)

    @property
    def time(self):
        """
        the current time : `Time`
        """
        return self.history.current_time

    @property
    def previous_time(self):
        """
        the time that has occurred last : `Time`
        """
        return self.history.previous_time


class Time(types.SimpleNamespace):
    """
    Represents an instance of time, having values for each :class:`TimeScale`

    Attributes
    ----------
        life : int : 0
            the `TimeScale.LIFE` value

        environment_sequence : int : 0
            the `TimeScale.ENVIRONMENT_SEQUENCE` value

        environment_state_update : int : 0
            the `TimeScale.ENVIRONMENT_STATE_UPDATE` value

        pass_ : int : 0
            the `TimeScale.PASS` value

        consideration_set_execution : int : 0
            the `TimeScale.CONSIDERATION_SET_EXECUTION` value

        absolute : `pint.Quantity` : 0ms
            the absolute time value

        absolute_interval : `pint.Quantity` : 1ms
            the interval between units of absolute time

        absolute_time_unit_scale : `TimeScale` : TimeScale.CONSIDERATION_SET_EXECUTION
            the `TimeScale` that corresponds to an interval of absolute time

        absolute_enabled : bool : False
            whether absolute time is used for this Time object

    """
    def __init__(
        self,
        consideration_set_execution=0,
        pass_=0,
        environment_state_update=0,
        environment_sequence=0,
        life=0,
        absolute=0 * _unit_registry.ms,
        absolute_interval=1 * _unit_registry.ms,
        absolute_time_unit_scale=TimeScale.CONSIDERATION_SET_EXECUTION,
        absolute_enabled=False,
        **alias_time_values,
    ):
        time_scale_values = {}
        for ts in TimeScale:
            time_scale_values[ts] = locals()[_time_scale_to_attr_str(ts)]

        time_scale_values.update({
            ts: alias_time_values[_time_scale_aliases[ts].lower()]
            for ts in _time_scale_aliases
            if _time_scale_aliases[ts].lower() in alias_time_values
        })

        super().__init__(
            **{
                _time_scale_to_attr_str(ts): v for ts, v in time_scale_values.items()
            },
            absolute=0 * _unit_registry.ms,
            absolute_interval=1 * _unit_registry.ms,
            absolute_time_unit_scale=TimeScale.CONSIDERATION_SET_EXECUTION,
            absolute_enabled=False,
        )

    def __repr__(self):
        abs_str = f'{self.absolute}, ' if self.absolute_enabled else ''
        ts_str = self._time_repr(exclusions={TimeScale.LIFE})
        return f'Time({abs_str}{ts_str})'

    def _time_repr(self, exclusions=()):
        ts_strs = []

        for ts in sorted(TimeScale, reverse=True):
            if ts not in exclusions:
                if ts not in _time_scale_aliases:
                    name = ts.name.lower()
                else:
                    name = _time_scale_aliases[ts].lower()

                ts_strs.append(
                    f'{name}: {getattr(self, _time_scale_to_attr_str(ts))}'
                )

        return ', '.join(ts_strs)

    def _get_by_time_scale(self, time_scale):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Returns
        -------
            this Time's value of a TimeScale by the TimeScale enum, rather \
            than by attribute : int
        """
        return getattr(self, _time_scale_to_attr_str(time_scale))

    def _set_by_time_scale(self, time_scale, value):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Sets this Time's value of a **time_scale** by the TimeScale enum,
        rather than by attribute
        """
        setattr(self, _time_scale_to_attr_str(time_scale), value)

    def _increment_by_time_scale(self, time_scale):
        """
        Increments the value of **time_scale** in this Time by one
        """
        self._set_by_time_scale(time_scale, self._get_by_time_scale(time_scale) + 1)
        self._reset_by_time_scale(time_scale)

        if (
            self.absolute_enabled
            and time_scale is self.absolute_time_unit_scale
        ):
            self.absolute += self.absolute_interval
            if (
                self.absolute_interval < 1 * self.absolute.u
                and isinstance(self.absolute.m, float)
            ):
                # only round for floating point interval increases,
                # until pint fixes precision errors
                # see https://github.com/hgrecco/pint/issues/1263
                self.absolute = round(self.absolute, _unit_registry.precision)

    def _reset_by_time_scale(self, time_scale):
        """
        Resets all the times for the time scale scope up to **time_scale**
        e.g. _reset_by_time_scale(TimeScale.ENVIRONMENT_STATE_UPDATE) will set the values for
        TimeScale.PASS and TimeScale.CONSIDERATION_SET_EXECUTION to 0
        """
        for relative_time_scale in sorted(TimeScale):
            # this works because the enum is set so that higher granularities of time have lower values
            if relative_time_scale >= time_scale:
                continue

            self._set_by_time_scale(relative_time_scale, 0)


class SimpleTime:
    """
    A subset class of `Time`, used to provide simple access to only
    `environment_sequence <Time.environment_sequence>`, `environment_state_update <Time.environment_state_update>`, and `consideration_set_execution <Time.consideration_set_execution>`
    """
    def __init__(self, time_ref):
        self._time_ref = time_ref

    # override __repr__ because this class is used only for cosmetic simplicity
    # based on a Time object
    def __repr__(self):
        abs_str = f'{self.absolute}, ' if self._time_ref.absolute_enabled else ''
        ts_str = self._time_ref._time_repr(exclusions={TimeScale.LIFE, TimeScale.PASS})
        return f'Time({abs_str}{ts_str})'

    @property
    def absolute(self):
        return self._time_ref.absolute

    @property
    def environment_sequence(self):
        return self._time_ref.environment_sequence

    @property
    def environment_state_update(self):
        return self._time_ref.environment_state_update

    @property
    def consideration_set_execution(self):
        return self._time_ref.consideration_set_execution


class TimeHistoryTree:
    """
    A tree object that stores a history of time that has occurred at various
    :class:`TimeScale`\\ s, typically used in conjunction with a `Clock`

    Attributes
    ----------
        time_scale : :class:`TimeScale` : `TimeScale.LIFE`
            the TimeScale unit this tree/node represents

        child_time_scale : :class:`TimeScale` : `TimeScale.ENVIRONMENT_SEQUENCE`
            the TimeScale unit for this tree's children

        children : list[`TimeHistoryTree`]
            an ordered list of this tree's children

        max_depth : :class:`TimeScale` : `TimeScale.ENVIRONMENT_STATE_UPDATE`
            the finest grain TimeScale that should be created as a subtree
            Setting this value lower allows for more precise measurements
            (by default, you cannot query the number of
            `TimeScale.CONSIDERATION_SET_EXECUTION`\\ s in a certain `TimeScale.PASS`), but
            this may use a large amount of memory in large simulations

        index : int
            the index this tree has in its parent's children list

        parent : `TimeHistoryTree` : None
            the parent node of this tree, if it exists. \
            None represents no parent (i.e. root node)

        previous_time : `Time`
            a `Time` object that represents the last time that has occurred in the tree

        current_time : `Time`
            a `Time` object that represents the current time in the tree

        total_times : dict{:class:`TimeScale`: int}
            stores the total number of units of :class:`TimeScale`\\ s that have \
            occurred over this tree's scope. Only contains entries for \
            :class:`TimeScale`\\ s of finer grain than **time_scale**

    Arguments
    ---------
        enable_current_time : bool : True
            sets this tree to maintain a `Time` object. If this tree is not
            a root (i.e. **time_scale** is `TimeScale.LIFE`)
    """
    def __init__(
        self,
        time_scale=TimeScale.LIFE,
        max_depth=TimeScale.ENVIRONMENT_STATE_UPDATE,
        index=0,
        parent=None,
        enable_current_time=True
    ):
        if enable_current_time:
            self.current_time = Time()
            self.previous_time = None
        self.index = index
        self.time_scale = time_scale
        self.max_depth = max_depth
        self.parent = parent

        self.child_time_scale = TimeScale.get_child(time_scale)

        if self.child_time_scale >= max_depth:
            self.children = [
                TimeHistoryTree(
                    self.child_time_scale,
                    max_depth=max_depth,
                    index=0,
                    parent=self,
                    enable_current_time=False
                )
            ]
        else:
            self.children = []

        self.total_times = {ts: 0 for ts in TimeScale if ts < self.time_scale}

    def increment_time(self, time_scale):
        """
        Increases this tree's **current_time** by one **time_scale**

        Arguments
        ---------
            time_scale : :class:`TimeScale`
                the unit of time to increment
        """
        if self.child_time_scale >= self.max_depth:
            if time_scale == self.child_time_scale:
                self.children.append(
                    TimeHistoryTree(
                        self.child_time_scale,
                        max_depth=self.max_depth,
                        index=len(self.children),
                        parent=self,
                        enable_current_time=False
                    )
                )
            else:
                self.children[-1].increment_time(time_scale)
        self.total_times[time_scale] += 1
        try:
            self.previous_time = copy.copy(self.current_time)
            self.current_time._increment_by_time_scale(time_scale)
        except AttributeError:
            # not all of these objects have time tracking
            pass

    def get_total_times_relative(
        self,
        query_time_scale,
        base_indices=None
    ):
        """
        Arguments
        ---------
            query_time_scale : :class:`TimeScale`
                the :class:`TimeScale` of units to be returned

            base_indices : dict{:class:`TimeScale`: int}
                a dictionary specifying what scope of time query_time_scale \
                is over. e.g.

                    base_indices = {TimeScale.ENVIRONMENT_SEQUENCE: 1, TimeScale.ENVIRONMENT_STATE_UPDATE: 5}

                gives the number of **query_time_scale**\\ s that have occurred \
                in the 5th `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` of the 1st `ENVIRONMENT_SEQUENCE`. If an entry for a :class:`TimeScale` \
                is not specified but is coarser than **query_time_scale**, the latest \
                value for that entry will be used

        Returns
        -------
            the number of units of query_time_scale that have occurred within \
            the scope of time specified by base_indices : int
        """
        if query_time_scale >= self.time_scale:
            raise TimeScaleError(
                'query_time_scale (given: {0}) must be of finer grain than {1}.time_scale ({2})'.format(
                    query_time_scale, self, self.time_scale
                )
            )

        try:
            self.current_time
        except AttributeError:
            raise TimeScaleError(
                'get_total_times_relative should only be called on a TimeHistoryTree with enable_current_time set to True'
            )

        default_base_indices = {
            TimeScale.LIFE: 0,
            TimeScale.ENVIRONMENT_SEQUENCE: None,
            TimeScale.ENVIRONMENT_STATE_UPDATE: None,
            TimeScale.PASS: None,
        }

        # overwrite defaults with dictionary passed in argument
        if base_indices is None:
            base_indices = default_base_indices
        else:
            default_base_indices.update(base_indices)
            base_indices = default_base_indices

        base_time_scale = TimeScale.LIFE
        # base_time_scale is set as the finest grain TimeScale that is specified,
        # but more coarse than query_time_scale
        # this will be where the query to attribute times will be made
        for ts in sorted(base_indices, reverse=True):
            if base_indices[ts] is not None and ts > query_time_scale:
                base_time_scale = ts

        if base_time_scale > self.time_scale:
            raise TimeScaleError(
                'base TimeScale set by base_indices ({0}) must be at least as fine as this TimeHistoryTree\'s time_scale ({1})'.format(
                    base_time_scale,
                    self.time_scale
                )
            )

        # get the root node, which will (and should) have TimeScale.LIFE
        node = self
        while node.parent is not None:
            node = node.parent

        try:
            # for all non-specified (i.e. set to None) TimeScales coarser than base_time_scale,
            # assign them to their latest time values as default
            while node.time_scale > base_time_scale:
                if base_indices[node.child_time_scale] is None:
                    base_indices[node.child_time_scale] = len(node.children) - 1
                node = node.children[base_indices[node.child_time_scale]]

            # attempt to retrieve the correct time count given the base_indices dictionary
            node = self
            while node.time_scale != base_time_scale:
                node = node.children[base_indices[node.child_time_scale]]
            return node.total_times[query_time_scale]
        except IndexError:
            raise TimeScaleError(
                'TimeHistoryTree {0}: {1} {2} does not exist in {3} {4}'.format(
                    self,
                    node.child_time_scale,
                    base_indices[node.child_time_scale],
                    node.time_scale,
                    node.index
                )
            )


@functools.lru_cache(maxsize=None)
def _time_scale_to_attr_str(time_scale: TimeScale) -> str:
    attr = time_scale.name.lower()

    if keyword.iskeyword(attr):
        return f'{attr}_'
    else:
        return attr


@functools.lru_cache(maxsize=None)
def _time_scale_to_class_str(time_scale: typing.Union[TimeScale, str]) -> str:
    try:
        name = time_scale.name
    except AttributeError:
        name = time_scale

    return ''.join([f'{x[0].upper()}{x[1:].lower()}' for x in name.split('_')])


def _attr_str_to_time_scale(attr_str):
    return getattr(TimeScale, attr_str.rstrip('_'))


def _multi_substitute_docstring(cls, subs: typing.Dict[str, str]):
    new_docstring = cls.__doc__
    try:
        for prev, new in subs.items():
            new_docstring = re.sub(prev, new, new_docstring)
    except TypeError:
        pass
    else:
        cls._old_docstring = cls.__doc__
        cls.__doc__ = new_docstring


def set_time_scale_alias(name: str, target: TimeScale):
    """Sets an alias named **name** of TimeScale **target**

    Args:
        name (str): name of the alias
        target (TimeScale): TimeScale that **name** will refer to
    """
    import graph_scheduler

    name_aliased_time_scales = list(filter(
        lambda e: _time_scale_aliases[e] == name,
        _time_scale_aliases
    ))
    if len(name_aliased_time_scales) > 0:
        raise ValueError(f"'{name}' is already aliased to {name_aliased_time_scales[0]}")

    try:
        target = getattr(TimeScale, target)
    except TypeError:
        pass
    except AttributeError as e:
        raise ValueError(f'Invalid TimeScale {target}') from e

    _time_scale_aliases[target] = name
    setattr(TimeScale, name, target)

    def getter(self):
        return getattr(self, _time_scale_to_attr_str(target))

    def setter(self, value):
        setattr(self, _time_scale_to_attr_str(target), value)

    prop = property(getter).setter(setter)
    setattr(Time, name.lower(), prop)
    setattr(SimpleTime, name.lower(), prop)

    # alias name in style of a class name
    new_class_segment_name = _time_scale_to_class_str(name)
    for cls_name, cls in graph_scheduler.__dict__.copy().items():
        # make aliases of conditions that contain a TimeScale name (e.g. AtEnvironmentStateUpdate)
        target_class_segment_name = _time_scale_to_class_str(target)

        if isinstance(cls, (type, types.ModuleType)):
            if isinstance(cls, types.ModuleType):
                try:
                    if _alias_docs_warning_str not in cls.__doc__:
                        cls.__doc__ = f'{_alias_docs_warning_str}{cls.__doc__}'
                except TypeError:
                    pass

            _multi_substitute_docstring(
                cls,
                {
                    target.name: name,
                    target_class_segment_name: new_class_segment_name,
                }
            )

        if target_class_segment_name in cls_name:
            new_cls_name = cls_name.replace(
                target_class_segment_name,
                new_class_segment_name
            )

            setattr(graph_scheduler.condition, new_cls_name, cls)
            setattr(graph_scheduler, new_cls_name, cls)

            graph_scheduler.condition.__all__.append(new_cls_name)
            graph_scheduler.__all__.append(new_cls_name)


def remove_time_scale_alias(name: str):
    """Removes an alias previously set by `set_time_scale_alias`

    Args:
        name (str): name of the TimeScale alias to remove
    """
    import graph_scheduler

    name_aliased_time_scales = list(filter(
        lambda e: _time_scale_aliases[e] == name,
        _time_scale_aliases
    ))
    assert len(name_aliased_time_scales) <= 1

    try:
        del _time_scale_aliases[name_aliased_time_scales[0]]
    except (IndexError, KeyError):
        return
    else:
        delattr(TimeScale, name)
        delattr(Time, name.lower())
        delattr(SimpleTime, name.lower())

        new_class_segment_name = _time_scale_to_class_str(name)
        for cls_name, cls in graph_scheduler.__dict__.copy().items():

            if isinstance(cls, (type, types.ModuleType)):
                if isinstance(cls, types.ModuleType):
                    if len(_time_scale_aliases) == 0:
                        try:
                            re.sub(_alias_docs_warning_str, '', cls.__doc__)
                        except TypeError:
                            pass

                try:
                    cls.__doc__ = cls._old_docstring
                    del cls._old_docstring
                except AttributeError:
                    # not all modules have docstrings
                    pass

            # NOTE: this could accidentally remove some unintended conditions if
            # an alias is constructed such that its determined class name is
            # a substring of some other Condition class name, but this is unlikely
            if new_class_segment_name in cls_name:
                delattr(graph_scheduler.condition, cls_name)
                delattr(graph_scheduler, cls_name)

                graph_scheduler.condition.__all__.remove(cls_name)
                graph_scheduler.__all__.remove(cls_name)



import os

import setuptools
import versioneer


here = os.path.abspath(os.path.dirname(__file__))


def get_requirements(require_name=None):
    prefix = require_name + '_' if require_name is not None else ''
    with open(os.path.join(here, prefix + 'requirements.txt'), encoding='utf-8') as f:
        return f.read().strip().split('\n')


setuptools.setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=get_requirements(),
    extras_require={
        'dev': get_requirements('dev'),
        'docs': get_requirements('docs'),
    }
)
